# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""The `airbyte_cdk.manifest_server.cli` module provides a standalone CLI for the Airbyte CDK Manifest Server.

This CLI enables running a FastAPI server for managing and executing Airbyte declarative manifests.
"""
